//
//  m014c9163spi.c
//  Graphic LCD module "itendo M014C9163SPI" I/O library
//  Copyright(C)2012 Toyohiko Togashi tog001@nifty.com
//
//
//  This program is free software; you can redistribute it and/or modify it under the terms of the
//  GNU General Public License as published by the Free Software Foundation; either version 3
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along with this program.
//  If not, see <http://www.gnu.org/licenses/>
//
//
//    Update history
//    ---------- ----- -------------------------------------------
//    2012.09.17 v0.0  First cording
//

#include <avr/io.h>						// WinAVR library
#include <avr/pgmspace.h>				// Flash access
#include <avr/interrupt.h>				// 
#include <util/delay.h>					// 

#include "m014c9163spi.h"				//

#define ABS(a)		(a < 0 ? -a : a)
#define OFFSET_X	0
#define OFFSET_Y	0

//
// 	Port I/O
//
void lcd_sendCmdData(unsigned char dc, unsigned char data) {
	unsigned char	i;
//	unsigned char	s;

//	s = SREG;
//	cli();

	if (dc) {
		LCD_PIN_DC_ON();
	} else {
		LCD_PIN_DC_OFF();
	}
	LCD_PIN_CS_ON();

#ifdef LCD_SPI_ENABLE					// SPI module mode
	SPDR = data;
	while(!(SPSR & _BV(SPIF)));
#else									// Software bit control mode
	for(i = 0; i < 8; i++) {
		if (data & 0x80) {
			LCD_PIN_SDA_ON();
		} else {
			LCD_PIN_SDA_OFF();
		}
		LCD_PIN_SCK_ON();
		data = data << 1;
		LCD_PIN_SCK_OFF();
	}
#endif

	LCD_PIN_CS_OFF();



//	if (s & _BV(SREG_I)) {				// Resume global interrupt
//		sei();
//	}

	return;
}

//
//	Open
//

void lcd_start(void) {
#ifdef LCD_PIN_RESET_INIT
	LCD_PIN_RESET_INIT();
#endif
	LCD_PIN_CS_INIT();
	LCD_PIN_DC_INIT();
#ifdef LCD_SPI_ENABLE					// SPI module mode
	DDRB  |=  _BV(DDB5);				// MOSI output
	PORTB &= ~_BV(PB5);					// MOSI low level
	DDRB  &= ~_BV(DDB4);				// MISO input
	PORTB &= ~_BV(PB5);					// Disable pull-up
	DDRB  |=  _BV(DDB3);				// SCK output
	PORTB &= ~_BV(PB5);					// SCK low level
//	DDRB  |=  _BV(DDB2);				// SS output
//	PORTB |=  _BV(PB2);					// SS high level
	SPCR =								// SPI Control Register
		//	_BV(SPIE) |					// SPI Interrupt Enable
			_BV(SPE)  |					// SPI Enable
		//	_BV(DORD) |					// Data Order   1:LSB 0:MSB
			_BV(MSTR) |					// Master/Slave Select  1:Master 0:Slave
		//	_BV(CPOL) |					// Clock Polarity         0  0  1  1
		//	_BV(CPHA) |					// Clock Phase            0  1  0  1
										//                        +--|--|--|-- SPI Mode 0 <==
										//                           +--|--|-- SPI Mode 1
										//                              +--|-- SPI Mode 2
										//                                 +-- SPI Mode 3
		//	_BV(SPI2X) <- SPSR			// Double SPI Speed Bit   0  0  0  0  1  1  1  1
		//	_BV(SPR1) |					// SPI Clock Rate Select  0  0  1  1  0  0  1  1
		//	_BV(SPR0) |					//                        0  1  0  1  0  1  0  1
										//                        +--|--|--|--|--|--|--|-- 1/4
										//                           +--|--|--|--|--|--|-- 1/16
										//                              +--|--|--|--|--|-- 1/64
										//                                 +--|--|--|--|-- 1/128
										//                                    +--|--|--|-- 1/2  <==
										//                                       +--|--|-- 1/8
										//                                          +--|-- 1/32
										//                                             +-- 1/64
			0;
	SPSR =								// SPI Status Register
		//	_BV(SPIF)  |				// SPI Interrupt Flag
		//	_BV(WCOL)  |				// Write COLlision Flag
			_BV(SPI2X) |				// Double SPI Speed Bit -> See SPCR
			0;
#else									// Software bit control mode
	LCD_PIN_SDA_INIT();					//
	LCD_PIN_SCK_INIT();					//
#endif

	return;
}

void lcd_init(void) {
	// Reset sequence
#ifdef LCD_PIN_RESET_ON
	LCD_PIN_RESET_OFF();
	_delay_ms(120);
	LCD_PIN_RESET_ON();
	_delay_ms(1);
	LCD_PIN_RESET_OFF();
#else
	lcd_sendCmdData(0, LCD_CMD_SOFT_RESET);
#endif
	_delay_ms(120);

	// Init code
	lcd_sendCmdData(0, LCD_CMD_EXIT_SLEEP_MODED);
	lcd_sendCmdData(0, LCD_CMD_SET_DISPLAY_ON);
	lcd_sendCmdData(0, LCD_CMD_SET_ADDRESS_MODE);
	lcd_sendCmdData(1, 0b11001000);		//  Row-rev, Column-rev, BGR order
#ifdef LCD_COLOR_16BIT
	lcd_sendCmdData(0, LCD_CMD_SET_PIXEL_FORMAT);
	lcd_sendCmdData(1, 0b01010101);		// 16bit/pixel
#endif

	// Screen clear
	lcd_draw_rect(0, 0, LCD_SCREEN_WIDTH, LCD_SCREEN_HEIGHT, LCD_COLOR_BLACK);

	return;
}

//
//	Close
//
void lcd_stop(void) {
#ifdef LCD_SPI_ENABLE					// SPI module mode
	SPCR  = 0;							// Reset SPI module
	DDRB  &= ~_BV(DDB5);				// Reset port (MOSI input)
	PORTB &= ~_BV(PB5);					// Reset port (disable pull-up)
	DDRB  &= ~_BV(DDB4);				// Reset port (MISO input)
	PORTB &= ~_BV(PB4);					// Reset port (disable pull-up)
	DDRB  &= ~_BV(DDB3);				// Reset port (SCK input)
	PORTB &= ~_BV(PB3);					// Reset port (disable pull-up)
#else									// Software bit control mode
	LCD_PIN_SCK_TERM();
	LCD_PIN_SDA_TERM();
#endif
	LCD_PIN_DC_TERM();
	LCD_PIN_CS_TERM();
#ifdef LCD_PIN_RESET_TERM
	LCD_PIN_RESET_TERM();
#endif
	return;
}

//
//	Rectangle
//
#ifdef LCD_COLOR_16BIT
const prog_uchar	lcd_color_code[][2] = {
//     rrrrrggg    gggbbbbb
	{0b00000000, 0b00000000},			// LCD_COLOR_BLACK		0
	{0b11111000, 0b00000000},			// LCD_COLOR_RED		1
	{0b11111000, 0b00000111},			// LCD_COLOR_PINK		2
	{0b11111000, 0b00001111},			// LCD_COLOR_PURPLE		3
	{0b11111000, 0b00011111},			// LCD_COLOR_MAGENTA	4
	{0b00000000, 0b00011111},			// LCD_COLOR_BLUE		5
	{0b00000111, 0b11111111},			// LCD_COLOR_CYAN		6
	{0b00000111, 0b11100000},			// LCD_COLOR_GREEN		7
	{0b01111111, 0b11100000},			// LCD_COLOR_LIME		8
	{0b11111111, 0b11100000},			// LCD_COLOR_YELLOW		9
	{0b11111011, 0b11100000},			// LCD_COLOR_ORANGE		10
	{0b01111011, 0b11101111},			// LCD_COLOR_GRAY		11
	{0b11111111, 0b11111111},			// LCD_COLOR_WHITE		12
};
#else
const prog_uchar	lcd_color_code[][3] = {
//   Red Green Blue
	{  0,   0,   0},					// LCD_COLOR_BLACK		0
	{255,   0,   0},					// LCD_COLOR_RED		1
	{255,   0,  64},					// LCD_COLOR_PINK		2
	{255,   0, 128},					// LCD_COLOR_PURPLE		3
	{255,   0, 255},					// LCD_COLOR_MAGENTA	4
	{  0,   0, 255},					// LCD_COLOR_BLUE		5
	{  0, 255, 255},					// LCD_COLOR_CYAN		6
	{  0, 255,   0},					// LCD_COLOR_GREEN		7
	{128, 255,   0},					// LCD_COLOR_LIME		8
	{255, 255,   0},					// LCD_COLOR_YELLOW		9
	{255, 128,   0},					// LCD_COLOR_ORANGE		10
	{128, 128, 128},					// LCD_COLOR_GRAY		11
	{255, 255, 255},					// LCD_COLOR_WHITE		12
};
#endif
void lcd_draw_rect(
	unsigned char	x,					// Upper right position
	unsigned char	y,					//
	unsigned char	w,					// width
	unsigned char	h,					// height
	unsigned char	color				//
) {
	short	i;
	short	m;
	unsigned char	r;
	unsigned char	g;
#ifndef LCD_COLOR_16BIT
	unsigned char	b;
#endif

	x += OFFSET_X;
	y += OFFSET_Y;

	r = pgm_read_byte(&lcd_color_code[color][0]);
	g = pgm_read_byte(&lcd_color_code[color][1]);
#ifndef LCD_COLOR_16BIT
	b = pgm_read_byte(&lcd_color_code[color][2]);
#endif
	lcd_sendCmdData(0, LCD_CMD_SET_PAGE_ADDRESS);	lcd_sendCmdData(1, 0);	lcd_sendCmdData(1, y);	lcd_sendCmdData(1, 0);	lcd_sendCmdData(1, y+h-1);
	lcd_sendCmdData(0, LCD_CMD_SET_COLUMN_ADDRESS);	lcd_sendCmdData(1, 0);	lcd_sendCmdData(1, x);	lcd_sendCmdData(1, 0);	lcd_sendCmdData(1, x+w-1);
	lcd_sendCmdData(0, LCD_CMD_WRITE_MEMORY_START);
	m = (short)h * (short)w;
	for(i = 0; i < m; i++) {
		lcd_sendCmdData(1, r);
		lcd_sendCmdData(1, g);
#ifndef LCD_COLOR_16BIT
		lcd_sendCmdData(1, b);
#endif
	}
	return;
}


//
//	Line
//
void lcd_draw_line(
	unsigned char	x1,					// Start point
	unsigned char	y1,					//
	unsigned char	x2,					// End point
	unsigned char	y2,					//
	unsigned char	s,					// Line size
	unsigned char	color				//
) {
	short			w;
	short			h;
	short			i;

	w = (short)x2 - (short)x1 + 1;
	h = (short)y2 - (short)y1 + 1;
//	if (w != 0) {
//		w++;
//	}
//	if (h != 0) {
//		h++;
//	}

	if (ABS(w) < ABS(h)) {				// 45..90deg
		if (h < 0) {
			for(i = 0; i > h; i--) {
				unsigned char	x,y;

				x = x1 + (i * w / h);
				y = y1 + i;
				lcd_draw_rect(x, y, s, s, color);
			}
		} else {
			for(i = 0; i < h; i++) {
				unsigned char	x,y;

				x = x1 + (i * w / h);
				y = y1 + i;
				lcd_draw_rect(x, y, s, s, color);
			}
		}
	} else {							// 0..45deg
		if (w < 0) {
			for(i = 0; i > w; i--) {
				unsigned char	x,y;

				x = x1 + i;
				y = y1 + (i * h / w);
				lcd_draw_rect(x, y, s, s, color);
			}
		} else {
			for(i = 0; i < w; i++) {
				unsigned char	x,y;

				x = x1 + i;
				y = y1 + (i * h / w);
				lcd_draw_rect(x, y, s, s, color);
			}
		}
	}

	return;
}

//
//	Character font
//
const prog_uchar	lcd_font5x7[][8] = {
	{0b00000,	// " "
	 0b00000,
	 0b00000,
	 0b00000,
	 0b00000,
	 0b00000,
	 0b00000,
	 0},
	{0b00100,	// !
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00000,
	 0b00100,
	 0},
	{0b01010,	// "
	 0b01010,
	 0b00000,
	 0b00000,
	 0b00000,
	 0b00000,
	 0b00000,
	 0},
	{0b01010,	// #
	 0b01010,
	 0b11111,
	 0b01010,
	 0b01010,
	 0b11111,
	 0b01010,
	 0},
	{0b00100,	// $
	 0b01111,
	 0b10100,
	 0b01110,
	 0b00101,
	 0b11111,
	 0b00100,
	 0},
	{0b01001,	// %
	 0b10101,
	 0b01010,
	 0b00100,
	 0b01010,
	 0b10101,
	 0b10010,
	 0},
	{0b01100,	// &
	 0b10010,
	 0b10100,
	 0b01000,
	 0b10101,
	 0b10010,
	 0b01101,
	 0},
	{0b00100,	// '
	 0b00100,
	 0b00000,
	 0b00000,
	 0b00000,
	 0b00000,
	 0b00000,
	 0},
	{0b00010,	// (
	 0b00100,
	 0b01000,
	 0b01000,
	 0b01000,
	 0b00100,
	 0b00010,
	 0},
	{0b01000,	// )
	 0b00100,
	 0b00010,
	 0b00010,
	 0b00010,
	 0b00100,
	 0b01000,
	 0},
	{0b00000,	// *
	 0b00100,
	 0b10101,
	 0b01110,
	 0b10101,
	 0b00100,
	 0b00000,
	 0},
	{0b00000,	// +
	 0b00100,
	 0b00100,
	 0b11111,
	 0b00100,
	 0b00100,
	 0b00000,
	 0},
	{0b00000,	// ,
	 0b00000,
	 0b00000,
	 0b00000,
	 0b01100,
	 0b00100,
	 0b01000,
	 0},
	{0b00000,	// -
	 0b00000,
	 0b00000,
	 0b11111,
	 0b00000,
	 0b00000,
	 0b00000,
	 0},
	{0b00000,	// .
	 0b00000,
	 0b00000,
	 0b00000,
	 0b00000,
	 0b01100,
	 0b01100,
	 0},
	{0b00000,	// /
	 0b00001,
	 0b00010,
	 0b00100,
	 0b01000,
	 0b10000,
	 0b00000,
	 0},
	{0b01110,	// 0
	 0b10001,
	 0b10001,
	 0b10001,
	 0b10001,
	 0b10001,
	 0b01110,
	 0},
	{0b00100,	// 1
	 0b01100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0},
	{0b01110,	// 2
	 0b10001,
	 0b00001,
	 0b00010,
	 0b00100,
	 0b01000,
	 0b11111,
	 0},
	{0b01110,	// 3
	 0b10001,
	 0b00001,
	 0b00110,
	 0b00001,
	 0b10001,
	 0b01110,
	 0},
	{0b10010,	// 4
	 0b10010,
	 0b10010,
	 0b11111,
	 0b00010,
	 0b00010,
	 0b00010,
	 0},
	{0b11111,	// 5
	 0b10000,
	 0b11110,
	 0b00001,
	 0b00001,
	 0b10001,
	 0b01110,
	 0},
	{0b01110,	// 6
	 0b10001,
	 0b10000,
	 0b11110,
	 0b10001,
	 0b10001,
	 0b01110,
	 0},
	{0b11111,	// 7
	 0b00001,
	 0b00010,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0},
	{0b01110,	// 8
	 0b10001,
	 0b10001,
	 0b01110,
	 0b10001,
	 0b10001,
	 0b01110,
	 0},
	{0b01110,	// 9
	 0b10001,
	 0b10001,
	 0b01111,
	 0b00001,
	 0b10001,
	 0b01110,
	 0},
	{0b00000,	// :
	 0b01100,
	 0b01100,
	 0b00000,
	 0b01100,
	 0b01100,
	 0b00000,
	 0},
	{0b00000,	// ;
	 0b01100,
	 0b01100,
	 0b00000,
	 0b01100,
	 0b00100,
	 0b01000,
	 0},
	{0b00010,	// <
	 0b00100,
	 0b01000,
	 0b10000,
	 0b01000,
	 0b00100,
	 0b00010,
	 0},
	{0b00000,	// =
	 0b00000,
	 0b11111,
	 0b00000,
	 0b11111,
	 0b00000,
	 0b00000,
	 0},
	{0b01000,	// >
	 0b00100,
	 0b00010,
	 0b00001,
	 0b00010,
	 0b00100,
	 0b01000,
	 0},
	{0b01110,	// ?
	 0b10001,
	 0b00001,
	 0b00010,
	 0b00100,
	 0b00000,
	 0b00100,
	 0},
	{0b01110,	// @
	 0b10001,
	 0b10111,
	 0b10101,
	 0b10111,
	 0b10000,
	 0b01110,
	 0},
	{0b00100,	// A
	 0b01010,
	 0b10001,
	 0b10001,
	 0b11111,
	 0b10001,
	 0b10001,
	 0},
	{0b11110,	// B
	 0b10001,
	 0b10001,
	 0b11110,
	 0b10001,
	 0b10001,
	 0b11110,
	 0},
	{0b01110,	// C
	 0b10001,
	 0b10000,
	 0b10000,
	 0b10000,
	 0b10001,
	 0b01110,
	 0},
	{0b11110,	// D
	 0b10001,
	 0b10001,
	 0b10001,
	 0b10001,
	 0b10001,
	 0b11110,
	 0},
	{0b11111,	// E
	 0b10000,
	 0b10000,
	 0b11111,
	 0b10000,
	 0b10000,
	 0b11111,
	 0},
	{0b11111,	// F
	 0b10000,
	 0b10000,
	 0b11111,
	 0b10000,
	 0b10000,
	 0b10000,
	 0},
	{0b01110,	// G
	 0b10001,
	 0b10000,
	 0b10000,
	 0b10011,
	 0b10001,
	 0b01110,
	 0},
	{0b10001,	// H
	 0b10001,
	 0b10001,
	 0b11111,
	 0b10001,
	 0b10001,
	 0b10001,
	 0},
	{0b01110,	// I
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b01110,
	 0},
	{0b00010,	// J
	 0b00010,
	 0b00010,
	 0b00010,
	 0b00010,
	 0b10010,
	 0b01100,
	 0},
	{0b10001,	// K
	 0b10010,
	 0b10100,
	 0b11000,
	 0b10100,
	 0b10010,
	 0b10001,
	 0},
	{0b10000,	// L
	 0b10000,
	 0b10000,
	 0b10000,
	 0b10000,
	 0b10000,
	 0b11111,
	 0},
	{0b10001,	// M
	 0b11011,
	 0b10101,
	 0b10001,
	 0b10001,
	 0b10001,
	 0b10001,
	 0},
	{0b10001,	// N
	 0b10001,
	 0b11001,
	 0b10101,
	 0b10011,
	 0b10001,
	 0b10001,
	 0},
	{0b01110,	// O
	 0b10001,
	 0b10001,
	 0b10001,
	 0b10001,
	 0b10001,
	 0b01110,
	 0},
	{0b11110,	// P
	 0b10001,
	 0b10001,
	 0b11110,
	 0b10000,
	 0b10000,
	 0b10000,
	 0},
	{0b01110,	// Q
	 0b10001,
	 0b10001,
	 0b10001,
	 0b10101,
	 0b10010,
	 0b01101,
	 0},
	{0b11110,	// R
	 0b10001,
	 0b10001,
	 0b11110,
	 0b10100,
	 0b10010,
	 0b10001,
	 0},
	{0b01110,	// S
	 0b10001,
	 0b10000,
	 0b01110,
	 0b00001,
	 0b10001,
	 0b01110,
	 0},
	{0b11111,	// T
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0},
	{0b10001,	// U
	 0b10001,
	 0b10001,
	 0b10001,
	 0b10001,
	 0b10001,
	 0b01110,
	 0},
	{0b10001,	// V
	 0b10001,
	 0b10001,
	 0b10001,
	 0b10001,
	 0b01010,
	 0b00100,
	 0},
	{0b10001,	// W
	 0b10001,
	 0b10001,
	 0b10101,
	 0b10101,
	 0b10101,
	 0b01010,
	 0},
	{0b10001,	// X
	 0b10001,
	 0b01010,
	 0b00100,
	 0b01010,
	 0b10001,
	 0b10001,
	 0},
	{0b10001,	// Y
	 0b01010,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0},
	{0b11111,	// Z
	 0b00001,
	 0b00010,
	 0b00100,
	 0b01000,
	 0b10000,
	 0b11111,
	 0},
	{0b01110,	// [
	 0b01000,
	 0b01000,
	 0b01000,
	 0b01000,
	 0b01000,
	 0b01110,
	 0},
	{0b00000,	// 
	 0b10000,
	 0b01000,
	 0b00100,
	 0b00010,
	 0b00001,
	 0b00000,
	 0},
	{0b01110,	// ]
	 0b00010,
	 0b00010,
	 0b00010,
	 0b00010,
	 0b00010,
	 0b01110,
	 0},
	{0b00100,	// ^
	 0b01010,
	 0b00000,
	 0b00000,
	 0b00000,
	 0b00000,
	 0b00000,
	 0},
	{0b00000,	// _
	 0b00000,
	 0b00000,
	 0b00000,
	 0b00000,
	 0b00000,
	 0b11111,
	 0},
	{0b01000,	// `
	 0b00100,
	 0b00000,
	 0b00000,
	 0b00000,
	 0b00000,
	 0b00000,
	 0},
	{0b00000,	// a
	 0b00000,
	 0b01110,
	 0b00001,
	 0b01111,
	 0b10011,
	 0b01101,
	 0},
	{0b10000,	// b
	 0b10000,
	 0b10000,
	 0b11110,
	 0b10001,
	 0b10001,
	 0b11110,
	 0},
	{0b00000,	// c
	 0b00000,
	 0b00000,
	 0b01111,
	 0b10000,
	 0b10000,
	 0b01111,
	 0},
	{0b00001,	// d
	 0b00001,
	 0b00001,
	 0b01111,
	 0b10001,
	 0b10001,
	 0b01111,
	 0},
	{0b00000,	// e
	 0b00000,
	 0b01110,
	 0b10001,
	 0b11111,
	 0b10000,
	 0b01110,
	 0},
	{0b00011,	// f
	 0b00100,
	 0b00100,
	 0b01110,
	 0b00100,
	 0b00100,
	 0b00100,
	 0},
	{0b00000,	// g
	 0b00000,
	 0b01110,
	 0b10001,
	 0b01111,
	 0b00001,
	 0b01110,
	 0},
	{0b10000,	// h
	 0b10000,
	 0b10000,
	 0b11110,
	 0b10001,
	 0b10001,
	 0b10001,
	 0},
	{0b00100,	// i
	 0b00000,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0},
	{0b00100,	// j
	 0b00000,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b11000,
	 0},
	{0b10000,	// k
	 0b10000,
	 0b10001,
	 0b10010,
	 0b10100,
	 0b11010,
	 0b10001,
	 0},
	{0b00100,	// l
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00110,
	 0},
	{0b00000,	// m
	 0b00000,
	 0b11010,
	 0b10101,
	 0b10101,
	 0b10001,
	 0b10001,
	 0},
	{0b00000,	// n
	 0b00000,
	 0b11110,
	 0b10001,
	 0b10001,
	 0b10001,
	 0b10001,
	 0},
	{0b00000,	// o
	 0b00000,
	 0b01110,
	 0b10001,
	 0b10001,
	 0b10001,
	 0b01110,
	 0},
	{0b00000,	// p
	 0b00000,
	 0b11110,
	 0b10001,
	 0b11110,
	 0b10000,
	 0b10000,
	 0},
	{0b00000,	// q
	 0b00000,
	 0b01111,
	 0b10001,
	 0b01111,
	 0b00001,
	 0b00001,
	 0},
	{0b00000,	// r
	 0b00000,
	 0b10111,
	 0b11000,
	 0b10000,
	 0b10000,
	 0b10000,
	 0},
	{0b00000,	// s
	 0b00000,
	 0b01110,
	 0b10000,
	 0b01110,
	 0b00001,
	 0b01110,
	 0},
	{0b00000,	// t
	 0b00100,
	 0b01110,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00010,
	 0},
	{0b00000,	// u
	 0b00000,
	 0b10001,
	 0b10001,
	 0b10001,
	 0b10001,
	 0b01111,
	 0},
	{0b00000,	// v
	 0b00000,
	 0b10001,
	 0b10001,
	 0b10001,
	 0b01010,
	 0b00100,
	 0},
	{0b00000,	// w
	 0b00000,
	 0b10001,
	 0b10001,
	 0b10101,
	 0b10101,
	 0b01010,
	 0},
	{0b00000,	// x
	 0b00000,
	 0b10001,
	 0b01010,
	 0b00100,
	 0b01010,
	 0b10001,
	 0},
	{0b00000,	// y
	 0b00000,
	 0b10001,
	 0b01010,
	 0b00100,
	 0b01000,
	 0b10000,
	 0},
	{0b00000,	// z
	 0b00000,
	 0b11111,
	 0b00010,
	 0b00100,
	 0b01000,
	 0b11111,
	 0},
	{0b00010,	// {
	 0b00100,
	 0b00100,
	 0b01000,
	 0b00100,
	 0b00100,
	 0b00010,
	 0},
	{0b00100,	// |
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0},
	{0b01000,	// }
	 0b00100,
	 0b00100,
	 0b00010,
	 0b00100,
	 0b00100,
	 0b01000,
	 0},
	{0b00000,	// ~
	 0b00000,
	 0b01000,
	 0b10101,
	 0b00010,
	 0b00000,
	 0b00000,
	 0},
	{0b00100,	// deg
	 0b01010,
	 0b00100,
	 0b00000,
	 0b00000,
	 0b00000,
	 0b00000,
	 0}
};

//
//	Text
//
void lcd_draw_char(unsigned char x, unsigned char y, unsigned char fcolor, unsigned char bcolor, char c) {
	unsigned char	i;
	unsigned char	j;
	unsigned char	fr;
	unsigned char	fg;
#ifndef LCD_COLOR_16BIT
	unsigned char	fb;
#endif
	unsigned char	br;
	unsigned char	bg;
#ifndef LCD_COLOR_16BIT
	unsigned char	bb;
#endif

	x += OFFSET_X;
	y += OFFSET_Y;
	c &= 0x7f;
	c -= ' ';
	fr = pgm_read_byte(&lcd_color_code[fcolor][0]);
	fg = pgm_read_byte(&lcd_color_code[fcolor][1]);
#ifndef LCD_COLOR_16BIT
	fb = pgm_read_byte(&lcd_color_code[fcolor][2]);
#endif
	br = pgm_read_byte(&lcd_color_code[bcolor][0]);
	bg = pgm_read_byte(&lcd_color_code[bcolor][1]);
#ifndef LCD_COLOR_16BIT
	bb = pgm_read_byte(&lcd_color_code[bcolor][2]);
#endif

	lcd_sendCmdData(0, LCD_CMD_SET_PAGE_ADDRESS);	lcd_sendCmdData(1, 0);	lcd_sendCmdData(1, y);	lcd_sendCmdData(1, 0);	lcd_sendCmdData(1, y+LCD_FONT_HEIGHT-1);
	lcd_sendCmdData(0, LCD_CMD_SET_COLUMN_ADDRESS);	lcd_sendCmdData(1, 0);	lcd_sendCmdData(1, x);	lcd_sendCmdData(1, 0);	lcd_sendCmdData(1, x+LCD_FONT_WIDTH-1);
	lcd_sendCmdData(0, LCD_CMD_WRITE_MEMORY_START);

	for(i = 0; i < LCD_FONT_HEIGHT; i++) {
		unsigned char	b;

		b = pgm_read_byte(&lcd_font5x7[c][i]);
		b <<= (8 - LCD_FONT_WIDTH);
		for(j = 0; j < LCD_FONT_WIDTH; j++) {
			b <<= 1;
			if (b & 0x80) {
				lcd_sendCmdData(1, fr);
				lcd_sendCmdData(1, fg);
#ifndef LCD_COLOR_16BIT
				lcd_sendCmdData(1, fb);
#endif
			} else {
				lcd_sendCmdData(1, br);
				lcd_sendCmdData(1, bg);
#ifndef LCD_COLOR_16BIT
				lcd_sendCmdData(1, bb);
#endif
			}
		}
	}
	return;
}

void lcd_draw_str(
	unsigned char	sx,					// Upper right position
	unsigned char	sy,					//
	unsigned char	fcolor,				// Text color
	unsigned char	bcolor,				// Background color
	char			*str				// '\0' terminated text
) {
	unsigned char	x;
	unsigned char	y;
	char			c;

	x  = sx;
	y  = sy;

	while((c = *str++) != '\0') {

		c &= 0b01111111;

		// Visual char
		if (c >= ' ') {
			lcd_draw_char(x, y, fcolor, bcolor, c);
			x += LCD_FONT_WIDTH;

		// CLF
		} else if (c == '\n') {
			x = sx;
			y += LCD_FONT_HEIGHT;

		// CR
		} else if (c == '\r') {
			y += LCD_FONT_HEIGHT;

		}

		// Screen range out
		if (x >= (LCD_SCREEN_WIDTH - LCD_FONT_WIDTH + 1)) {
			x = sx;
			y += LCD_FONT_HEIGHT;
		}
		if (y >= (LCD_SCREEN_HEIGHT - LCD_FONT_HEIGHT + 1)) {
			y = 0;
		}
	}

	return;
}

//
//	Virtual character display interface
//
#ifdef LCD_VCD
unsigned char	lcd_vcd_x1 = 0;
unsigned char	lcd_vcd_y1 = 0;
unsigned char	lcd_vcd_x2 = LCD_SCREEN_WIDTH  - LCD_FONT_WIDTH  + 1;
unsigned char	lcd_vcd_y2 = LCD_SCREEN_HEIGHT - LCD_FONT_HEIGHT + 1;
unsigned char	lcd_vcd_w  = LCD_SCREEN_WIDTH;
unsigned char	lcd_vcd_h  = LCD_SCREEN_HEIGHT;
unsigned char	lcd_curs_x = 0;
unsigned char	lcd_curs_y = 0;
unsigned char	lcd_curs_fcolor = LCD_COLOR_WHITE;
unsigned char	lcd_curs_bcolor = LCD_COLOR_BLACK;

void lcd_setViewPort(unsigned char	x, unsigned char y, unsigned char w, unsigned char h) {
	lcd_vcd_x1 = x;
	lcd_vcd_y1 = y;
	lcd_vcd_x2 = x + w - LCD_FONT_WIDTH  + 1;
	lcd_vcd_y2 = y + h - LCD_FONT_HEIGHT + 1;
	lcd_vcd_w  = w;
	lcd_vcd_h  = h;
	lcd_curs_x = x;
	lcd_curs_y = y;
	return;
}

void lcd_position(unsigned char	column, unsigned char row) {
	lcd_curs_x = column * LCD_FONT_WIDTH  + lcd_vcd_x1;
	lcd_curs_y = row    * LCD_FONT_HEIGHT + lcd_vcd_y1;
	return;
}

void lcd_color(unsigned char fcolor) {
	lcd_curs_fcolor = fcolor;
	return;
}

void lcd_colors(unsigned char fcolor, unsigned char bcolor) {
	lcd_curs_fcolor = fcolor;
	lcd_curs_bcolor = bcolor;
	return;
}

void lcd_prString(char *str) {
	char	c;

	while((c = *str++) != '\0') {

		c &= 0b01111111;

		// Visual char
		if (c >= ' ') {
			lcd_draw_char(lcd_curs_x, lcd_curs_y, lcd_curs_fcolor, lcd_curs_bcolor, c);
			lcd_curs_x += LCD_FONT_WIDTH;

		// LF
		} else if (c == '\n') {
			lcd_curs_x = lcd_vcd_x1;
			lcd_curs_y += LCD_FONT_HEIGHT;

		// CR
		} else if (c == '\r') {
			lcd_curs_x = lcd_vcd_x1;

		// FF
		} else if (c == '\f') {
			lcd_control(LCD_DISP_CLEAR_HOME);

		// VT
		} else if (c == '\v') {
			lcd_control(LCD_CURSOR_HOME);
		}

		// View range out
		if (lcd_curs_x >= lcd_vcd_x2) {
			lcd_curs_x = lcd_vcd_x1;
			lcd_curs_y += LCD_FONT_HEIGHT;
		}
		if (lcd_curs_y >= lcd_vcd_y2) {
			lcd_curs_y = lcd_vcd_y1;
		}
	}

	return;
}

void lcd_control(unsigned char cmd) {
	switch(cmd) {
	case LCD_DISP_CLEAR_HOME:
		lcd_draw_rect(
			lcd_vcd_x1, lcd_vcd_y1,
			lcd_vcd_w, lcd_vcd_h,
			lcd_curs_bcolor);
	case LCD_CURSOR_HOME:
		lcd_curs_x = lcd_vcd_x1;
		lcd_curs_y = lcd_vcd_y1;
		break;
	default:
		break;
	}
	return;
}

#endif	// LCD_VCD


