//
//  m014c9163spi.h
//  Graphic LCD module "itendo M014C9163SPI" I/O library
//  Copyright(C)2012 Toyohiko Togashi tog001@nifty.com
//
//
//  This program is free software; you can redistribute it and/or modify it under the terms of the
//  GNU General Public License as published by the Free Software Foundation; either version 3
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along with this program.
//  If not, see <http://www.gnu.org/licenses/>
//
//
//	Description
//		I/O control for Graphic LCD module.
//		LCD module: itendo M014C9163SPI. (Controller chip is ILITEK ILI9163)
//		Interface: SPI+1bit by software control.
//			D/C  : Any output port (LCD_PIN_DC)
//			MOSI : Any output port (LCD_PIN_SDA) ...SPI
//			MISO : Not used
//			SCLK : Any output port (LCD_PIN_SCK) ...SPI
//			CS   : Any output port (LCD_PIN_CS) ...SPI
//			RESET: Any output port (LCD_PIN_RESET) ...optinal
//		Screen direction:
//			X: 0=Left .. 127=Right
//			Y: 0=Top  .. 127=Bottom
//		Color mode: RGB6-6-6bits or RGB5-6-5bits
//
//  Device: ATmega328P       IDE: AVR Studio 4.18.700, WinAVR 20100110(GCC 4.3.3)
//
//
//    Update history
//    ---------- ----- -------------------------------------------
//    2012.09.17 v0.0  First cording
//    2012.09.25 v0.0  Support AVR interval SPI module  (!! Not work)
//    2012.09.30 v0.0  Support 16bit color mode
//


//------------------------------------------------------------------------------
// Hardware config & Option
//------------------------------------------------------------------------------
//#define LCD_SPI_ENABLE				// Not use, not work

#define LCD_PIN_DC_ON()			(PORTC |=  _BV(PORTC0))
#define LCD_PIN_DC_OFF()		(PORTC &= ~_BV(PORTC0))
#define LCD_PIN_DC_INIT()		(DDRC  |=  _BV(DDC0), LCD_PIN_DC_OFF())
#define LCD_PIN_DC_TERM()		(DDRC  &= ~_BV(DDC0), PORTC &= ~_BV(PORTC0))

#define LCD_PIN_SDA_ON()		(PORTB |=  _BV(PORTB3))
#define LCD_PIN_SDA_OFF()		(PORTB &= ~_BV(PORTB3))
#define LCD_PIN_SDA_INIT()		(DDRB  |=  _BV(DDB3), LCD_PIN_SDA_OFF())
#define LCD_PIN_SDA_TERM()		(DDRB  &= ~_BV(DDB3), PORTB &= ~_BV(PORTB3))

#define LCD_PIN_SCK_ON()		(PORTB |=  _BV(PORTB5))
#define LCD_PIN_SCK_OFF()		(PORTB &= ~_BV(PORTB5))
#define LCD_PIN_SCK_INIT()		(DDRB  |=  _BV(DDB5), LCD_PIN_SCK_OFF())
#define LCD_PIN_SCK_TERM()		(DDRB  &= ~_BV(DDB5), PORTB &= ~_BV(PORTB5))

#define LCD_PIN_CS_ON()			(PORTC &= ~_BV(PORTC3))		// Neg
#define LCD_PIN_CS_OFF()		(PORTC |=  _BV(PORTC3))		// Neg
#define LCD_PIN_CS_INIT()		(DDRC  |=  _BV(DDC3), LCD_PIN_CS_OFF())
#define LCD_PIN_CS_TERM()		(DDRC  &= ~_BV(DDC3), PORTC &= ~_BV(PORTC3))

#define LCD_PIN_RESET_ON()		(PORTC &= ~_BV(PORTC4))		// Neg
#define LCD_PIN_RESET_OFF()		(PORTC |=  _BV(PORTC4))		// Neg
#define LCD_PIN_RESET_INIT()	(DDRC  |=  _BV(DDC4), LCD_PIN_RESET_OFF())
#define LCD_PIN_RESET_TERM()	(DDRC  &= ~_BV(DDC4), PORTC &= ~_BV(PORTC4))


#define LCD_COLOR_16BIT					// Red5-Green6-Blue5 Mode
//#define LCD_COLOR_24BIT				// Red8-Green8-Blue8 Mode

#define LCD_VCD							// Enable VCD(Virtual Charactor Display)
//#define F_CPU 8000000UL				// Please IDE project configuration option frequency[Hz]


//------------------------------------------------------------------------------
// Layer 1
//------------------------------------------------------------------------------

extern void lcd_start(void);
extern void lcd_init(void);
extern void lcd_stop(void);
extern void lcd_sendCmdData(unsigned char data_or_command, unsigned char data) ;
#define LCD_DC_COMMAND		0
#define LCD_DC_DATA			1
// LCD module native command code
#define LCD_CMD_SOFT_RESET				0x01	// None
#define LCD_CMD_ENTER_SLEEP_MODE		0x10	// None
#define LCD_CMD_EXIT_SLEEP_MODED		0x11	// None
#define LCD_CMD_ENTER_PARTIAL_MODE		0x12	// None
#define LCD_CMD_ENTER_NORMAL_MODE		0x13	// None
#define LCD_CMD_EXIT_INVERT_MODE		0x20	// None
#define LCD_CMD_ENTER_INVERT_MODE		0x21	// None
#define LCD_CMD_SET_GAMMA_CURVE			0x26	// 1byte-W 
#define LCD_CMD_SET_DISPLAY_OFF			0x28	// None 
#define LCD_CMD_SET_DISPLAY_ON			0x29	// None 
#define LCD_CMD_SET_COLUMN_ADDRESS		0x2a	// 4byte-W
#define LCD_CMD_SET_PAGE_ADDRESS		0x2b	// 4byte-W
#define LCD_CMD_WRITE_MEMORY_START		0x2c	// Nbytes-W
#define LCD_CMD_WRITE_LUT				0x2d	// Nbytes-W
#define LCD_CMD_SET_PARTIAL_AREA		0x30	// 4byte
#define LCD_CMD_SET_SCROLL_AREA			0x33	// 6byte
#define LCD_CMD_SET_TEAR_OFF			0x34	// None
#define LCD_CMD_SET_TEAR_ON				0x35	// 1byte-W
#define LCD_CMD_SET_ADDRESS_MODE		0x36	// 1byte-W
#define LCD_CMD_SET_SCROLL_START		0x37	// 2byte-W
#define LCD_CMD_EXIT_IDLE_MODE			0x38	// None
#define LCD_CMD_ENTER_IDLE_MODE			0x39	// None
#define LCD_CMD_SET_PIXEL_FORMAT		0x3a	// 1byte-W
#define LCD_CMD_WRITE_MEMORY_CONTINUE	0x3c	// Nbytes-W
#define LCD_CMD_SET_TEAR_SCANLINE		0x44	// 2byte-W



//------------------------------------------------------------------------------
//	Layer 2
//------------------------------------------------------------------------------

//
// Prototype
//
#define     lcd_open()		(lcd_start(), lcd_init())
#define     lcd_close()		lcd_stop()
extern void lcd_draw_rect(unsigned char x, unsigned char y, unsigned char width, unsigned char height, unsigned char color);
extern void lcd_draw_line(unsigned char x1, unsigned char y1, unsigned char x2, unsigned char y2, unsigned char line_size, unsigned char color);
extern void lcd_draw_str(unsigned char x, unsigned char y, unsigned char face_color, unsigned char back_color, char *str);
#ifdef LCD_VCD
extern void lcd_setViewPort(unsigned char x, unsigned char y, unsigned char width, unsigned char height);
extern void lcd_prString(char *str);
#define     lcd_prCString(s)	lcd_prString(s)
extern void lcd_position(unsigned char column, unsigned char row);
extern void lcd_color(unsigned char face_color);
extern void lcd_colors(unsigned char face_color, unsigned char back_color);
extern void lcd_control(unsigned char command);
#endif

//
// General parameter
//
#define LCD_SCREEN_WIDTH	128
#define LCD_SCREEN_HEIGHT	128
#define LCD_FONT_WIDTH		6
#define LCD_FONT_HEIGHT		8

//
// Color code
//                          
#define LCD_COLOR_BLACK		0
#define LCD_COLOR_RED		1
#define LCD_COLOR_PINK		2
#define LCD_COLOR_PURPLE	3
#define LCD_COLOR_MAGENTA	4
#define LCD_COLOR_BLUE		5
#define LCD_COLOR_CYAN		6
#define LCD_COLOR_GREEN		7
#define LCD_COLOR_LIME		8
#define LCD_COLOR_YELLOW	9
#define LCD_COLOR_ORANGE	10
#define LCD_COLOR_GRAY		11
#define LCD_COLOR_WHITE		12

//
//	VCD lcd_control() command code
//
#ifdef LCD_VCD
#define LCD_DISP_ON                       0x0C
#define LCD_DISP_OFF                      0x08
#define LCD_DISP_BLANK                    0x0A
#define LCD_DISP_CLEAR_HOME               0x01
#define LCD_CURSOR_ON                     0x0E
#define LCD_CURSOR_OFF                    0x0C
#define LCD_CURSOR_WINK                   0x0D
#define LCD_CURSOR_BLINK                  0x0F
#define LCD_CURSOR_SH_LEFT                0x10
#define LCD_CURSOR_SH_RIGHT               0x14
#define LCD_CURSOR_HOME                   0x02
#define LCD_CURSOR_LEFT                   0x04
#define LCD_CURSOR_RIGHT                  0x06
#define LCD_PORT_MASK                     0x7F
#endif
